"""Implementation of common cast conversion.

Mainly following interfaces are defined.

- to_int_from_float
    Convert float value to int.
- to_float_from_int
    Convert int value to float.
- to_bool_from_int
    Convert int value to bool.
"""


from typing import Any
from typing import Union

import apysc as ap
from apysc._type.number_value_interface import NumberValueInterface


def to_int_from_float(
        int_or_float: Union[int, float, NumberValueInterface]) -> Any:
    """
    Convert float value to int.

    Parameters
    ----------
    int_or_float : int or float or Int or Number
        Target float value. If integer is specified, conversion
        will be skipped.

    Returns
    -------
    int_val : int or Int
        Converted integer value.
    """
    with ap.DebugInfo(
            callable_=to_int_from_float, locals_=locals(),
            module_name=__name__):
        if isinstance(int_or_float, ap.Number):
            return ap.Int(int_or_float)
        if not isinstance(int_or_float, float):
            return int_or_float
        return int(int_or_float)


def to_float_from_int(
        int_or_float: Union[int, float, NumberValueInterface]) -> Any:
    """
    Convert int value to float.

    Parameters
    ----------
    int_or_float : int or float or Int or Number
        Target integer value. If float is specified, conversion will
        be skipped.

    Returns
    -------
    float_val : float or Number
        Converted float value.
    """
    with ap.DebugInfo(
            callable_=to_float_from_int, locals_=locals(),
            module_name=__name__):
        if isinstance(int_or_float, ap.Int):
            return ap.Number(int_or_float)
        if not isinstance(int_or_float, int):
            return int_or_float
        return float(int_or_float)


def to_bool_from_int(integer: Union[int, ap.Int]) -> bool:
    """
    Convert int value to bool.

    Parameters
    ----------
    integer : int or Int
        Integer value to convert.

    Returns
    -------
    bool_val : bool
        Converted boolean value.

    Raises
    ------
    ValueError
        If argument value isn't zero or one.
    """
    with ap.DebugInfo(
            callable_=to_bool_from_int, locals_=locals(),
            module_name=__name__):
        from apysc._validation import number_validation
        number_validation.validate_int_is_zero_or_one(integer=integer)
        if integer == 0:
            return False
        return True



"""Each interface to get a Python built-in value from an apysc one.
"""

from typing import Union

import apysc as ap


def get_builtin_str_from_apysc_val(string: Union[str, ap.String]) -> str:
    """
    Get a Python built-in string from an apysc one.

    Parameters
    ----------
    string : str or ap.String
        Target string value.

    Returns
    -------
    builtin_val : str
        A Python built-in string.
    """
    with ap.DebugInfo(
            callable_=get_builtin_str_from_apysc_val, locals_=locals(),
            module_name=__name__):
        if isinstance(string, str):
            return string
        builtin_val: str = string._value
        return builtin_val



"""Class implementation for the animation_height interface.
"""

from typing import Union

import apysc as ap
from apysc._animation.animation_height import AnimationHeight
from apysc._animation.animation_interface_base import AnimationInterfaceBase
from apysc._animation.easing import Easing


class AnimationHeightInterface(AnimationInterfaceBase):

    def animation_height(
            self,
            height: Union[int, ap.Int],
            duration: Union[int, ap.Int] = 3000,
            delay: Union[int, ap.Int] = 0,
            easing: Easing = Easing.LINEAR) -> AnimationHeight:
        """
        Set the height animation setting.

        Notes
        -----
        To start this animation, you need to call the `start` method of
        the returned instance.

        Parameters
        ----------
        height : int or Int
            The final height of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_height : AnimationHeight
            Created animation setting instance.

        References
        ----------
        - animation_width and animation_height interfaces document
            - https://bit.ly/39XPUdq
        - Animation interfaces duration setting document
            - https://simon-ritchie.github.io/apysc/animation_duration.html
        - Animation interfaces delay setting document
            - https://simon-ritchie.github.io/apysc/animation_delay.html
        - Each animation interface return value document
            - https://bit.ly/2XOoa8w
        - Sequential animation setting document
            - https://simon-ritchie.github.io/apysc/sequential_animation.html
        - animation_parallel interface document
            - https://simon-ritchie.github.io/apysc/animation_parallel.html
        - Easing enum document
            - https://simon-ritchie.github.io/apysc/easing_enum.html
        """
        animation_height: AnimationHeight = AnimationHeight(
            target=self,
            height=height,
            duration=duration,
            delay=delay,
            easing=easing)
        return animation_height



"""Class implementation for graphics clear method related interface.
"""


from typing import Optional

import apysc as ap
from apysc._display.line_dash_dot_setting import LineDashDotSetting
from apysc._display.line_dash_setting import LineDashSetting
from apysc._display.line_dot_setting import LineDotSetting
from apysc._display.line_round_dot_setting import LineRoundDotSetting
from apysc._display.polyline import Polyline


class GraphicsClearInterface:

    _fill_color: ap.String
    _fill_alpha: ap.Number
    _line_color: ap.String
    _line_thickness: ap.Int
    _line_alpha: ap.Number
    _children: ap.Array
    _current_line: Optional[Polyline]
    _line_cap: ap.String
    _line_joints: ap.String
    _line_dot_setting: Optional[LineDotSetting]
    _line_dash_setting: Optional[LineDashSetting]
    _line_round_dot_setting: Optional[LineRoundDotSetting]
    _line_dash_dot_setting: Optional[LineDashDotSetting]

    def clear(self) -> None:
        """
        Clear all graphics and reset fill and line settings.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.clear, locals_=locals(),
                module_name=__name__, class_=GraphicsClearInterface):
            from apysc._display.begin_fill_interface import BeginFillInterface
            from apysc._display.child_interface import ChildInterface
            from apysc._display.fill_alpha_interface import FillAlphaInterface
            from apysc._display.fill_color_interface import FillColorInterface
            from apysc._display.line_alpha_interface import LineAlphaInterface
            from apysc._display.line_color_interface import LineColorInterface
            from apysc._display.line_style_interface import LineStyleInterface
            from apysc._display.line_thickness_interface import \
                LineThicknessInterface
            if isinstance(self, (FillColorInterface, BeginFillInterface)):
                self._initialize_fill_color_if_not_initialized()
            self._fill_color.value = ''
            if isinstance(self, (FillAlphaInterface, BeginFillInterface)):
                self._initialize_fill_alpha_if_not_initialized()
            self._fill_alpha.value = 1.0
            if isinstance(self, (LineColorInterface, LineStyleInterface)):
                self._initialize_line_color_if_not_initialized()
            self._line_color.value = ''
            if isinstance(self, (LineThicknessInterface, LineStyleInterface)):
                self._initialize_line_thickness_if_not_initialized()
            self._line_thickness.value = 1
            if isinstance(self, (LineAlphaInterface, LineStyleInterface)):
                self._initialize_line_alpha_if_not_initialized()
            self._line_alpha.value = 1.0
            if isinstance(self, ChildInterface):
                self._initialize_children_if_not_initialized()
            while self._children:
                self._children[0].remove_from_parent()
            if hasattr(self, '_current_line'):
                self._current_line = None
            if hasattr(self, '_line_cap'):
                self._line_cap = ap.String(ap.LineCaps.BUTT.value)
            if hasattr(self, '_line_joints'):
                self._line_joints = ap.String(ap.LineJoints.MITER.value)
            self._line_dot_setting = None
            self._line_dash_setting = None
            self._line_round_dot_setting = None
            self._line_dash_dot_setting = None



"""Class implementation for child related interface.
"""

from typing import Any
from typing import Dict
from typing import List
from typing import Union

import apysc as ap
from apysc._display.display_object import DisplayObject
from apysc._type.revert_interface import RevertInterface


class ChildInterface(RevertInterface):

    _children: ap.Array[DisplayObject]
    _variable_name: str
    stage: 'ap.Stage'

    def add_child(self, child: DisplayObject) -> None:
        """
        Add display object child to this instance.

        Parameters
        ----------
        child : DisplayObject
            Child instance to add.

        References
        ----------
        - Sprite class add_child and remove_child interfaces document
            - https://bit.ly/2Ugk47G
        """
        with ap.DebugInfo(
                callable_=self.add_child, locals_=locals(),
                module_name=__name__, class_=ChildInterface):
            from apysc._validation import display_validation
            self._initialize_children_if_not_initialized()
            display_validation.validate_display_object(display_object=child)
            if child.parent is not None:
                child.remove_from_parent()
            self._children.append(child)
            child.parent = self
            append_expression_of_add_child(child=child)

    def _initialize_children_if_not_initialized(self) -> None:
        """
        Initialize _children attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_children'):
            return
        self._children = ap.Array([])

    def remove_child(self, child: DisplayObject) -> None:
        """
        Remove display object child from this instance.

        Parameters
        ----------
        child : DisplayObject
            Child instance to remove.

        References
        ----------
        - Sprite class add_child and remove_child interfaces document
            - https://bit.ly/2Ugk47G
        """
        with ap.DebugInfo(
                callable_=self.remove_child, locals_=locals(),
                module_name=__name__, class_=ChildInterface):
            self._initialize_children_if_not_initialized()
            append_expression_of_remove_child(child=child)
            for child_ in self._children.value:  # type: ignore
                if child_ != child:
                    continue
                self._children.remove(child)
                child.parent = None
                return

    def contains(self, child: DisplayObject) -> ap.Boolean:
        """
        Get a boolean whether this instance contains specified child.

        Parameters
        ----------
        child : DisplayObject
            Child instance to check.

        Returns
        -------
        result : Boolean
            If this instance contains specified child, True will
            be set.

        References
        ----------
        - Sprite class contains interface document
            - https://simon-ritchie.github.io/apysc/sprite_contains.html
        """
        with ap.DebugInfo(
                callable_=self.contains, locals_=locals(),
                module_name=__name__, class_=ChildInterface):
            self._initialize_children_if_not_initialized()
            index: ap.Int = self._children.index_of(value=child)
            if index == -1:
                result: ap.Boolean = ap.Boolean(False)
            else:
                result = ap.Boolean(True)
            self._append_contains_expression(result=result, child=child)
            return result

    def _append_contains_expression(
            self, result: ap.Boolean, child: DisplayObject) -> None:
        """
        Append contains method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        child : DisplayObject
            Child instance to check.
        """
        expression: str = (
            f'{result.variable_name} = '
            f'{self._variable_name}.has({child.variable_name});'
        )
        ap.append_js_expression(expression=expression)

    @property
    def num_children(self) -> ap.Int:
        """
        Get a current children number.

        Returns
        -------
        num_children : int
            Current children number.

        References
        ----------
        - Sprite class num_children interface document
            - https://simon-ritchie.github.io/apysc/sprite_num_children.html
        """
        with ap.DebugInfo(
                callable_='num_children', locals_=locals(),
                module_name=__name__, class_=ChildInterface):
            self._initialize_children_if_not_initialized()
            num_children: ap.Int = ap.Int(value=self._children.length)
            self._append_num_children_expression(num_children=num_children)
            return num_children

    def _append_num_children_expression(
            self, num_children: ap.Int) -> None:
        """
        Append num_children method expression.

        Parameters
        ----------
        num_children : Int
            Current children number.
        """
        expression: str = (
            f'{num_children.variable_name} = '
            f'{self._variable_name}.children().length;'
        )
        ap.append_js_expression(expression=expression)

    def get_child_at(self, index: Union[int, ap.Int]) -> DisplayObject:
        """
        Get child at specified index.

        Parameters
        ----------
        index : int or Int
            Child's index (start from 0).

        Returns
        -------
        child : DisplayObject
            Target index child instance.

        References
        ----------
        - Sprite class get_child_at interface document
            - https://bit.ly/3rggoi6
        """
        with ap.DebugInfo(
                callable_=self.get_child_at, locals_=locals(),
                module_name=__name__, class_=ChildInterface):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            self._initialize_children_if_not_initialized()
            if self.num_children > index:
                child: DisplayObject = self._children[index]
            else:
                variable_name: str = expression_variables_util.\
                    get_next_variable_name(type_name=var_names.DISPLAY_OBJECT)
                child = DisplayObject(
                    stage=self.stage, variable_name=variable_name)
            self._append_get_child_at_expression(child=child, index=index)
            return child

    def _append_get_child_at_expression(
            self, child: DisplayObject, index: Union[int, ap.Int]) -> None:
        """
        Append get_child_at method expression.

        Parameters
        ----------
        child : DisplayObject
            Target index child instance.
        index : int or Int
            Child's index (start from 0).
        """
        from apysc._type import value_util
        index_str: str = value_util.get_value_str_for_expression(value=index)
        expression: str = (
            f'var {child.variable_name} = '
            f'{self._variable_name}.children()'
            f'[{index_str}];'
        )
        ap.append_js_expression(expression=expression)

    _children_snapshots: Dict[str, List[Any]]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make values' snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_children_snapshots'):
            self._children_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_children_if_not_initialized()

        for child in self._children.value:  # type: ignore
            if not isinstance(child, RevertInterface):
                continue
            child._run_all_make_snapshot_methods(snapshot_name=snapshot_name)

        self._children_snapshots[snapshot_name] = [*self._children._value]

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert values if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._children._value = [*self._children_snapshots[snapshot_name]]

        for child in self._children.value:  # type: ignore
            if not isinstance(child, RevertInterface):
                continue
            child._run_all_revert_methods(snapshot_name=snapshot_name)


def append_expression_of_add_child(child: DisplayObject) -> None:
    """
    Append expression of add_child (add).

    Parameters
    ----------
    child : DisplayObject
        Child object to add.
    """
    with ap.DebugInfo(
            callable_=append_expression_of_add_child, locals_=locals(),
            module_name=__name__):
        parent_name: str = child.parent.variable_name  # type: ignore
        child_name: str = child.variable_name
        expression: str = (
            f'{parent_name}.add({child_name});'
        )
        ap.append_js_expression(expression=expression)


def append_expression_of_remove_child(child: DisplayObject) -> None:
    """
    Append expression of remove_child (removeElement).

    Parameters
    ----------
    child : DisplayObject
        Child object to remove.
    """
    with ap.DebugInfo(
            callable_=append_expression_of_remove_child, locals_=locals(),
            module_name=__name__):
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        parent_name: str = expression_variables_util.get_next_variable_name(
            type_name=var_names.PARENT)
        child_name: str = child.variable_name
        expression: str = (
            f'var {parent_name} = {child_name}.parent();'
            f'\nif ({parent_name}) {{'
            f'\n  {parent_name}.removeElement({child_name});'
            '\n}'
        )
        ap.append_js_expression(expression=expression)



"""Event validation implementation.

Mainly following interfaces are defined.

- validate_event
    Validate specified instance is Event.
"""

from typing import Any

import apysc as ap


def validate_event(e: Any) -> ap.Event:
    """
    Validate whether specified instance is Event or not.

    Parameters
    ----------
    e : Event
        Event instance to check.

    Raises
    ------
    ValueError
        If specified instance is not Event instance.

    Returns
    -------
    e : Event
        Event instance.
    """
    if isinstance(e, ap.Event):
        return e
    raise ValueError(
        f'Specified instance is not Event type: {type(e)}')


def validate_event_type(mouse_event_type: Any) -> ap.MouseEventType:
    """
    Validate whether specified value is MouseEventType one or not.

    Parameters
    ----------
    mouse_event_type : MouseEventType
        EventTMouseEventTypeype value to check.

    Returns
    -------
    mouse_event_type : MouseEventType
        MouseEventType value.
    """
    if isinstance(mouse_event_type, ap.MouseEventType):
        return mouse_event_type
    raise ValueError(
        f'Specified value is not a MouseEventType: {type(mouse_event_type)}')



"""Class implementation for line joints interface.
"""

from typing import Dict
from typing import Union

import apysc as ap
from apysc._display.line_joints import LineJoints
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class LineJointsInterface(VariableNameInterface, RevertInterface):

    _line_joints: ap.String

    def _initialize_line_joints_if_not_initialized(self) -> None:
        """
        Initialize _line_joints attribute if that it is not
        initialized yet.
        """
        if hasattr(self, '_line_joints'):
            return
        self._line_joints = ap.String(LineJoints.MITER.value)

    @property
    def line_joints(self) -> Union[ap.String, LineJoints]:
        """
        Get this instance's line joints style setting.

        Returns
        -------
        line_joints : String
            Line joints style setting.
        """
        with ap.DebugInfo(
                callable_='line_joints', locals_=locals(),
                module_name=__name__, class_=LineJointsInterface):
            self._initialize_line_joints_if_not_initialized()
            return self._line_joints._copy()

    @line_joints.setter
    def line_joints(self, value: Union[ap.String, LineJoints]) -> None:
        """
        Set line joints style setting.

        Parameters
        ----------
        value : String or LineJoints
            Line joints style setting to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_joints', locals_=locals(),
                module_name=__name__, class_=LineJointsInterface):
            self._update_line_joints_and_skip_appending_exp(value=value)
            self._append_line_joints_update_expression()

    def _update_line_joints_and_skip_appending_exp(
            self, value: Union[ap.String, LineJoints]) -> None:
        """
        Update line joints and skip appending expression.

        Parameters
        ----------
        value : STring or LineJoints
            Line joints style setting to set.
        """
        from apysc._validation.display_validation import validate_line_joints
        if not isinstance(value, (ap.String, LineJoints)):
            raise TypeError(
                'Not supported line_joints type specified: '
                f'{type(value)}'
                '\nAcceptable ones are: String or LineJoints.')
        validate_line_joints(joints=value)
        if isinstance(value, ap.String):
            self._line_joints = value._copy()
        else:
            self._line_joints = ap.String(value.value)

    def _append_line_joints_update_expression(self) -> None:
        """
        Append line cap updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_line_joints_update_expression,
                locals_=locals(),
                module_name=__name__, class_=LineJointsInterface):
            from apysc._type import value_util
            joints_name: str = value_util.get_value_str_for_expression(
                value=self._line_joints)
            expression: str = (
                f'{self.variable_name}.attr'
                f'({{"stroke-linejoin": {joints_name}}});'
            )
            ap.append_js_expression(expression=expression)

    _line_joints_snapshots: Dict[str, str]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_line_joints_snapshots'):
            self._line_joints_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_line_joints_if_not_initialized()
        self._line_joints_snapshots[snapshot_name] = self._line_joints._value

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._line_joints._value = self._line_joints_snapshots[snapshot_name]



"""Class implementation for radius value interface.
"""

from typing import Dict
from typing import Union

import apysc as ap
from apysc._animation.animation_radius_interface import \
    AnimationRadiusInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.revert_interface import RevertInterface


class RadiusInterface(
        AnimationRadiusInterface, RevertInterface, AttrLinkingInterface):

    _radius: ap.Int

    def _initialize_radius_if_not_initialized(self) -> None:
        """
        Initialize _radius attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_radius'):
            return
        self._radius = ap.Int(0)

        self._append_raidus_attr_linking_setting()

    def _append_raidus_attr_linking_setting(self) -> None:
        """
        Append a radius attribute linking setting.
        """
        with ap.DebugInfo(
                callable_=self._append_raidus_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=RadiusInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._radius, attr_name='radius')
            self._append_attr_to_linking_stack(
                attr=self._radius, attr_name='radius')

    @property
    def radius(self) -> ap.Int:
        """
        Get radius value.

        Returns
        -------
        radius : Int
            Radius value.
        """
        with ap.DebugInfo(
                callable_='radius', locals_=locals(),
                module_name=__name__, class_=RadiusInterface):
            from apysc._type import value_util
            self._initialize_radius_if_not_initialized()
            return value_util.get_copy(value=self._radius)

    @radius.setter
    def radius(self, value: ap.Int) -> None:
        """
        Update radius value.

        Parameters
        ----------
        value : int or Int
            Radius value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='radius', locals_=locals(),
                module_name=__name__, class_=RadiusInterface):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=value)
            value = self._get_converted_radius_int(radius=value)
            self._radius = value
            self._radius._append_incremental_calc_substitution_expression()
            self._append_radius_update_expression()

            self._append_raidus_attr_linking_setting()

    def _get_converted_radius_int(self, radius: Union[int, ap.Int]) -> ap.Int:
        """
        Get a radius converted Int instance.

        Parameters
        ----------
        radius : int or Int
            Radius value.

        Returns
        -------
        radius : Int
            Type converted radius value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._get_converted_radius_int, locals_=locals(),
                module_name=__name__, class_=RadiusInterface):
            if not isinstance(radius, ap.Int):
                return ap.Int(radius)
            return radius

    def _append_radius_update_expression(self) -> None:
        """
        Append radius value updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_radius_update_expression,
                locals_=locals(),
                module_name=__name__, class_=RadiusInterface):
            from apysc._type import value_util
            self._initialize_radius_if_not_initialized()
            value_str: str = value_util.get_value_str_for_expression(
                value=self._radius)
            expression: str = (
                f'{self.variable_name}.radius({value_str});'
            )
            ap.append_js_expression(expression=expression)

    _radius_snapshots: Dict[str, int]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_radius_snapshots'):
            self._radius_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_radius_if_not_initialized()
        self._radius_snapshots[snapshot_name] = int(self._radius._value)

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._radius._value = self._radius_snapshots[snapshot_name]



"""Class implementation for the animation_skew_y interface.
"""

import warnings
from typing import Union

import apysc as ap
from apysc._animation.animation_interface_base import AnimationInterfaceBase
from apysc._animation.animation_skew_y import AnimationSkewY
from apysc._animation.easing import Easing


class AnimationSkewYInterface(AnimationInterfaceBase):

    def _animation_skew_y(
            self,
            skew_y: Union[int, ap.Int],
            duration: Union[int, ap.Int] = 3000,
            delay: Union[int, ap.Int] = 0,
            easing: Easing = Easing.LINEAR) -> AnimationSkewY:
        """
        **Important notes**
        Currently this interface does not work correctly.
        For more details please see:
        https://github.com/svgdotjs/svg.js/issues/1222

        Set the skew-y animation animation.

        Notes
        -----
        To start this animation, you need to call the `start` method of
        the returned instance.

        Parameters
        ----------
        skew_y : int or Int
            The final skew-y of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_skew_y : AnimationSkewY
            Created animation setting instance.

        References
        ----------
        - Animation interfaces duration setting document
            - https://simon-ritchie.github.io/apysc/animation_duration.html
        - Animation interfaces delay setting document
            - https://simon-ritchie.github.io/apysc/animation_delay.html
        - Each animation interface return value document
            - https://bit.ly/2XOoa8w
        - Sequential animation setting document
            - https://simon-ritchie.github.io/apysc/sequential_animation.html
        - animation_parallel interface document
            - https://simon-ritchie.github.io/apysc/animation_parallel.html
        - Easing enum document
            - https://simon-ritchie.github.io/apysc/easing_enum.html
        """
        warnings.warn(
            'Currently the `animation_skew_y` interface does not '
            'work correctly. Maybe this issue will be fixed in future '
            'versions. For more details, please see '
            'https://github.com/svgdotjs/svg.js/issues/1222')
        animation_skew_y: AnimationSkewY = AnimationSkewY(
            target=self,
            skew_y=skew_y,
            duration=duration,
            delay=delay,
            easing=easing)
        return animation_skew_y



"""Class implementation for dictionary.
"""

from typing import Any
from typing import Dict
from typing import Generic
from typing import TypeVar
from typing import Union

import apysc as ap
from apysc._event.custom_event_interface import CustomEventInterface
from apysc._type.copy_interface import CopyInterface
from apysc._type.dictionary_structure import DictionaryStructure
from apysc._type.expression_string import ExpressionString
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface

DefaultType = TypeVar('DefaultType')

_BuiltinKeys = Union[str, int, float]
_K = TypeVar(
    '_K', str, int, float, ap.String, ap.Int, ap.Number, ExpressionString)
_V = TypeVar('_V')


class Dictionary(
        Generic[_K, _V],
        CopyInterface, RevertInterface, DictionaryStructure,
        CustomEventInterface):
    """
    Dictionary class for the apysc library.
    """

    _initial_value: Union[Dict[_K, _V], 'Dictionary']
    _value: Dict[_K, _V]

    def __init__(self, value: Union[Dict[_K, _V], 'Dictionary']) -> None:
        """
        Dictionary class for the apysc library.

        Parameters
        ----------
        value : dict or Dictionary
            Initial dictionary value.
        """
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            with TemporaryNotHandlerScope():
                TYPE_NAME: str = var_names.DICTIONARY
                self._validate_acceptable_value_type(value=value)
                self._initial_value = value
                self._type_name = TYPE_NAME
                self._value = self._get_dict_value(value=value)
                self.variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=TYPE_NAME)
                self._append_constructor_expression()

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=self._initial_value)
            expression: str = f'var {self.variable_name} = {value_str};'
            ap.append_js_expression(expression=expression)

    def _get_dict_value(
            self, value: Union[Dict[_K, _V], 'Dictionary']) -> Dict[_K, _V]:
        """
        Get a dict value from specified value.

        Parameters
        ----------
        value : dict or Dictionary
            Specified dictionary value.

        Returns
        -------
        dict_val : dict
            Converted dict value.
        """
        if isinstance(value, Dictionary):
            return value._value
        return value

    def _validate_acceptable_value_type(
            self, value: Union[Dict[_K, _V], 'Dictionary']) -> None:
        """
        Validate that specified value is acceptable type or not.

        Parameters
        ----------
        value : dict or Dictionary
            Dictionary value to check.

        Raises
        ------
        TypeError
            If specified value's type is not dict or Dictionary.
        """
        if isinstance(value, (dict, Dictionary)):
            return
        raise TypeError(
            'Not acceptable value type is specified.'
            f'\nSpecified valkue type is: {type(value)}'
            '\nAcceptable types are: dict and Dictionary'
        )

    @property
    def value(self) -> Union[Dict[_K, _V], 'Dictionary']:
        """
        Get a current dict value.

        Returns
        -------
        value : dict
            Current dict value.

        References
        ----------
        - apysc basic data classes common value interface
            - https://bit.ly/3Be1aij
        """
        return self._value

    @value.setter
    def value(self, value: Union[Dict[_K, _V], 'Dictionary']) -> None:
        """
        Set dictionary value.

        Parameters
        ----------
        value : dict or Dictionary.
            Dictionary value to set.

        References
        ----------
        apysc basic data classes common value interface
            https://bit.ly/3Be1aij
        """
        with ap.DebugInfo(
                callable_='value', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            self._validate_acceptable_value_type(value=value)
            self._value = self._get_dict_value(value=value)
            self._append_value_setter_expression(value=value)

    def _append_value_setter_expression(
            self, value: Union[Dict[_K, _V], 'Dictionary']) -> None:
        """
        Append value's setter expression.

        Parameters
        ----------
        value : dict or Dictionary.
            Dictionary value to set.
        """
        with ap.DebugInfo(
                callable_=self._append_value_setter_expression,
                locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = f'{self.variable_name} = {value_str};'
            ap.append_js_expression(expression=expression)

    _value_snapshot: Dict[str, Dict[_K, _V]]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make values' snapthot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_value_snapshot'):
            self._value_snapshot = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value_snapshot[snapshot_name] = {**self._value}

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert values if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value = self._value_snapshot[snapshot_name]

    def __str__(self) -> str:
        """
        String conversion method.

        Returns
        -------
        string : str
            Converted value string.
        """
        if not hasattr(self, '_value'):
            return '{}'
        return str(self._value)

    def __repr__(self) -> str:
        """
        Get a representation string of this instance.

        Returns
        -------
        repr_str : str
            Representation string of this instance.
        """
        if not hasattr(self, '_value'):
            repr_str: str = 'Dictionary({})'
        else:
            repr_str = f'Dictionary({self._value})'
        return repr_str

    @property
    def length(self) -> ap.Int:
        """
        Get length of this dictionary values.

        Returns
        -------
        length : Int
            This dictionary value's length.
        """
        with ap.DebugInfo(
                callable_='length', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            length: ap.Int = ap.Int(len(self._value))
            self._append_length_expression(length=length)
            return length

    def _append_length_expression(self, length: ap.Int) -> None:
        """
        Append length method expression.

        Parameters
        ----------
        length : Int
            Created length Int variable.
        """
        with ap.DebugInfo(
                callable_=self._append_length_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            expression: str = (
                f'{length.variable_name} = '
                f'Object.keys({self.variable_name}).length;'
            )
            ap.append_js_expression(expression=expression)

    def __len__(self) -> None:
        """
        This method is disabled and can't use from Dictionary instance.
        """
        raise Exception(
            'Dictionary instance can\'t apply len function.'
            ' Please use length property instead.')

    def __getitem__(self, key: Union[_K, ExpressionString]) -> _V:
        """
        Get a specified key's single value.

        Parameters
        ----------
        key : _K
            Dictionary key.

        Returns
        -------
        value : *
            Specified key's value.
        """
        with ap.DebugInfo(
                callable_='__getitem__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            self._validate_key_type_is_str_or_numeric(key=key)
            key_: _BuiltinKeys = self._get_builtin_type_key(key=key)
            has_key: bool = key_ in self._value
            if has_key:
                value: Any = self._value[key_]  # type: ignore
            else:
                value = ap.AnyValue(None)
            self._append_getitem_expression(key=key, value=value)
            return value

    def _get_builtin_type_key(
            self, key: Union[_K, ExpressionString]) -> _BuiltinKeys:
        """
        Get a built-in type's key (str, int, or float) from
        a specified key.

        Parameters
        ----------
        key : _K
            Target key value (including String, Int, and Number).

        Returns
        -------
        key : str or int or float
            Built-in type's key.
        """
        if isinstance(key, ExpressionString):
            return key.value
        if isinstance(key, (ap.String, ap.Int, ap.Number)):
            return key._value
        return key

    def _append_getitem_expression(
            self, key: Union[_K, ExpressionString], value: Any) -> None:
        """
        Append __getitem__ expression.

        Parameters
        ----------
        key : _K
            Dictionary key.
        value : *
            Specified key's value.
        """
        with ap.DebugInfo(
                callable_=self._append_getitem_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            if not isinstance(value, VariableNameInterface):
                value = ap.AnyValue(None)
            key_str: str = value_util.get_value_str_for_expression(value=key)
            expression: str = (
                f'var {value.variable_name} = '
                f'{self.variable_name}[{key_str}];'
            )
            ap.append_js_expression(expression=expression)

    def _validate_key_type_is_str_or_numeric(
            self, key: Union[_K, ExpressionString]) -> None:
        """
        Validate whether key value type is acceptable (str or int or
        flaot) or not.

        Parameters
        ----------
        key : _K
            Dictionary key to check.

        Raises
        ------
        ValueError
            If key type is not str, String, int, Int, float, or Number.
        """
        if isinstance(
                key,
                (str, ap.String, int, ap.Int, float, ap.Number,
                 ExpressionString)):
            return
        raise ValueError(
            f'Unsupported key type is specified: {type(key)}, {key}'
            f'\nSuppoting types are: str, String, int, Int')

    def __setitem__(
            self, key: Union[_K, ExpressionString], value: _V) -> None:
        """
        Set value to a specified key position.

        Parameters
        ----------
        key : _K
            Dictionary key to set value.
        value : *
            Any value to set.
        """
        with ap.DebugInfo(
                callable_='__setitem__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            key_: _BuiltinKeys = self._get_builtin_type_key(key=key)
            self._value[key_] = value  # type: ignore
            self._append_setitem_expression(key=key, value=value)

    def _append_setitem_expression(
            self, key: Union[_K, ExpressionString], value: _V) -> None:
        """
        Append __setitem__ method expression.

        Parameters
        ----------
        key : _K
            Dictionary key to check.
        value : *
            Any value to set.
        """
        with ap.DebugInfo(
                callable_=self._append_setitem_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            key_str: str = value_util.get_value_str_for_expression(value=key)
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = (
                f'{self.variable_name}[{key_str}] = {value_str};'
            )
            ap.append_js_expression(expression=expression)

    def __delitem__(self, key: Union[_K, ExpressionString]) -> None:
        """
        Delete specified key's value from dictionary.

        Parameters
        ----------
        key : _K
            Dictionary key to delete.
        """
        with ap.DebugInfo(
                callable_='__delitem__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            key_: _BuiltinKeys = self._get_builtin_type_key(key=key)
            if key_ in self._value:
                del self._value[key_]  # type: ignore
            self._append_delitem_expression(key=key)

    def _append_delitem_expression(
            self, key: Union[_K, ExpressionString]) -> None:
        """
        Append __delitem__ method expression.

        Parameters
        ----------
        key : _K
            Dictionary key to delete.
        """
        with ap.DebugInfo(
                callable_=self._append_delitem_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            key_str: str = value_util.get_value_str_for_expression(value=key)
            expression: str = (
                f'delete {self.variable_name}[{key_str}];'
            )
            ap.append_js_expression(expression=expression)

    def __eq__(self, other: Any) -> Any:
        """
        Equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. dict or Dictionary types are acceptable.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        with ap.DebugInfo(
                callable_='__eq__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            if isinstance(other, Dictionary):
                result: ap.Boolean = ap.Boolean(self._value == other._value)
            else:
                result = ap.Boolean(self._value == other)
                if isinstance(other, dict):
                    other = Dictionary(other)
            if isinstance(other, VariableNameInterface):
                self._append_eq_expression(result=result, other=other)
            return result

    def _append_eq_expression(
            self, result: ap.Boolean,
            other: VariableNameInterface) -> None:
        """
        Append an __eq__ expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : Dictionary
            Dictionary other value to compare.
        """
        with ap.DebugInfo(
                callable_=self._append_eq_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            expression: str = (
                f'{result.variable_name} = '
                f'_.isEqual({self.variable_name}, {other.variable_name});'
            )
            ap.append_js_expression(expression=expression)

    def __ne__(self, other: Any) -> Any:
        """
        Noe equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. dict or Dictionary types are acceptable.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        with ap.DebugInfo(
                callable_='__ne__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            if isinstance(other, dict):
                other = Dictionary(other)
            result: ap.Boolean = self == other
            result = result.not_
            if isinstance(other, VariableNameInterface):
                self._append_ne_expression(result=result, other=other)
            return result

    def _append_ne_expression(
            self, result: ap.Boolean,
            other: VariableNameInterface) -> None:
        """
        Append a __ne__ expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : Dictionary
            Dictionary other value to compare.
        """
        with ap.DebugInfo(
                callable_=self._append_ne_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            expression: str = (
                f'{result.variable_name} = '
                f'!_.isEqual({self.variable_name}, {other.variable_name});'
            )
            ap.append_js_expression(expression=expression)

    def get(
            self, key: Union[_K, ExpressionString],
            default: DefaultType = None) -> DefaultType:
        """
        Get a specified key dictionary value. If this dictionary
        hasn't a specified key, then a default value will be returned.

        Parameters
        ----------
        key : _K
            Target key.
        default : DefaultType or None, optional
            Any default value.

        Returns
        -------
        result_value : Any
            Extracted value or a default value.
        """
        key_: _BuiltinKeys = self._get_builtin_type_key(key=key)
        if key_ in self._value:
            result_value: Any = self._value[key_]  # type: ignore
        else:
            result_value = default
        self._append_get_expression(
            key=key, result_value=result_value, default=default)
        return result_value

    def _append_get_expression(
            self, key: Union[_K, ExpressionString],
            result_value: Any,
            default: Any) -> None:
        """
        Append the `get` method expression.

        Parameters
        ----------
        key : _K
            Target key.
        result_value : Any
            Extracted value or a default value.
        default : Any
            Any default value. Basically apysc types (e.g., Int, Number,
            String, and so on) are necessary.
        """
        from apysc._type import value_util
        key_str: str = value_util.get_value_str_for_expression(value=key)
        if not isinstance(result_value, VariableNameInterface):
            result_value = ap.AnyValue(value=None)
        result_value_str: str = value_util.get_value_str_for_expression(
            value=result_value)
        default_value_str: str = value_util.get_value_str_for_expression(
            value=default)
        expression: str = (
            f'if ({key_str} in {self.variable_name}) {{'
            f'\n  {result_value_str} = {self.variable_name}[{key_str}];'
            '\n}else {'
            f'\n  {result_value_str} = {default_value_str};'
            '\n}'
        )
        ap.append_js_expression(expression=expression)



"""Class implementation for the parallel animation interface.
"""

from typing import List
from typing import Union

import apysc as ap
from apysc._animation.animation_base import AnimationBase
from apysc._animation.animation_interface_base import AnimationInterfaceBase
from apysc._animation.animation_parallel import AnimationParallel
from apysc._animation.easing import Easing


class AnimationParallelInterface(AnimationInterfaceBase):

    def animation_parallel(
            self,
            animations: List[AnimationBase],
            duration: Union[int, ap.Int] = 3000,
            delay: Union[int, ap.Int] = 0,
            easing: Easing = Easing.LINEAR) -> AnimationParallel:
        """
        Set the parallel animation setting.

        Notes
        -----
        - To start this animation, you need to call the `start` method of
            the returned instance.
        - The `animations` argument can't contains the `AnimationParallel`
            instance.
        - The `duration`, `delay`, and `easing` arguments in the
            `animations` argument will be ignored (this interface's
            arguments will be refered instead).

        Raises
        ------
        ValueError
            - If the animations's target is not this instance.
            - If there are changed `duration`, `delay`, or
                `easing` animation settings in the `animations`
                list.

        Parameters
        ----------
        animations : list of AnimationBase
            Target animation settings.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_parallel : AnimationParallel
            Created animation setting instance.

        References
        ----------
        - Animation interfaces duration setting document
            - https://simon-ritchie.github.io/apysc/animation_duration.html
        - Animation interfaces delay setting document
            - https://simon-ritchie.github.io/apysc/animation_delay.html
        - Each animation interface return value document
            - https://bit.ly/2XOoa8w
        - Sequential animation setting document
            - https://simon-ritchie.github.io/apysc/sequential_animation.html
        - animation_parallel interface document
            - https://simon-ritchie.github.io/apysc/animation_parallel.html
        - Easing enum document
            - https://simon-ritchie.github.io/apysc/easing_enum.html
        """
        animation_parallel: AnimationParallel = AnimationParallel(
            target=self,
            animations=animations,
            duration=duration,
            delay=delay,
            easing=easing)
        return animation_parallel



"""Class implementation for ellipse the width interface.
"""

from typing import Dict

import apysc as ap
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class EllipseWidthInterface(
        VariableNameInterface, RevertInterface, AttrLinkingInterface):

    _ellipse_width: ap.Int

    def _initialize_ellipse_width_if_not_initialized(self) -> None:
        """
        Initialize _ellipse_width attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_ellipse_width'):
            return
        self._ellipse_width = ap.Int(0)

        self._append_ellipse_width_attr_linking_setting()

    def _append_ellipse_width_attr_linking_setting(self) -> None:
        """
        Append a ellipse-height attribute linking setting.
        """
        with ap.DebugInfo(
                callable_=self._append_ellipse_width_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=EllipseWidthInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._ellipse_width, attr_name='ellipse_width')
            self._append_attr_to_linking_stack(
                attr=self._ellipse_width, attr_name='ellipse_width')

    @property
    def ellipse_width(self) -> ap.Int:
        """
        Get ellipse width value.

        Returns
        -------
        ellipse_width : Int
            Ellipse width value.
        """
        with ap.DebugInfo(
                callable_='ellipse_width', locals_=locals(),
                module_name=__name__, class_=EllipseWidthInterface):
            from apysc._type import value_util
            self._initialize_ellipse_width_if_not_initialized()
            return value_util.get_copy(value=self._ellipse_width)

    @ellipse_width.setter
    def ellipse_width(self, value: ap.Int) -> None:
        """
        Update ellipse width value.

        Parameters
        ----------
        value : int or Int
            Ellipse width value.
        """
        with ap.DebugInfo(
                callable_='ellipse_width', locals_=locals(),
                module_name=__name__, class_=EllipseWidthInterface):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=value)
            if isinstance(value, int):
                value = ap.Int(value)
            self._ellipse_width = value
            self._ellipse_width.\
                _append_incremental_calc_substitution_expression()
            self._append_ellipse_width_update_expression()

            self._append_ellipse_width_attr_linking_setting()

    def _append_ellipse_width_update_expression(self) -> None:
        """
        Append ellipse width updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ellipse_width_update_expression,
                locals_=locals(),
                module_name=__name__, class_=EllipseWidthInterface):
            from apysc._type import value_util
            self._initialize_ellipse_width_if_not_initialized()
            width_value_str: str = value_util.get_value_str_for_expression(
                value=self._ellipse_width)
            if hasattr(self, '_ellipse_height'):
                height_value_str: str = value_util.\
                    get_value_str_for_expression(
                        value=getattr(self, '_ellipse_height'))
            else:
                height_value_str = value_util.get_value_str_for_expression(
                    value=0)
            expression: str = (
                f'{self.variable_name}.radius({width_value_str}, '
                f'{height_value_str});'
            )
            ap.append_js_expression(expression=expression)

    _ellipse_width_snapshots: Dict[str, int]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make the value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_ellipse_width_snapshots'):
            self._ellipse_width_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_ellipse_width_if_not_initialized()
        self._ellipse_width_snapshots[snapshot_name] = int(
            self._ellipse_width._value)

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert the value if the snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._ellipse_width._value = self._ellipse_width_snapshots[
            snapshot_name]



"""Class implementation for the skew x interface.
"""

from typing import Dict

import apysc as ap
from apysc._animation.animation_skew_x_interface import AnimationSkewXInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.revert_interface import RevertInterface


class SkewXInterface(
        AnimationSkewXInterface, RevertInterface, AttrLinkingInterface):

    _skew_x: ap.Int

    def _initialize_skew_x_if_not_initialized(self) -> None:
        """
        Initialize the _skew_x attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_skew_x'):
            return
        self._skew_x = ap.Int(0)

        self._append_skew_x_attr_linking_setting()

    def _append_skew_x_attr_linking_setting(self) -> None:
        """
        Append a skew-x attribute linking setting.
        """
        with ap.DebugInfo(
                callable_='skew_x', locals_=locals(),
                module_name=__name__, class_=SkewXInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._skew_x, attr_name='skew_x')
            self._append_attr_to_linking_stack(
                attr=self._skew_x, attr_name='skew_x')

    @property
    def skew_x(self) -> ap.Int:
        """
        Get a current skew x value of the instance.

        Returns
        -------
        skew_x : Int
            Current skew x value of this instance.

        References
        ----------
        - GraphicsBase skew_x and skew_y interfaces document
            - https://simon-ritchie.github.io/apysc/graphics_base_skew.html
        """
        with ap.DebugInfo(
                callable_='skew_x', locals_=locals(),
                module_name=__name__, class_=SkewXInterface):
            from apysc._type import value_util
            self._initialize_skew_x_if_not_initialized()
            return value_util.get_copy(value=self._skew_x)

    @skew_x.setter
    def skew_x(self, value: ap.Int) -> None:
        """
        Update a skew x value of this instance.

        Parameters
        ----------
        value : Int
            Skew x value to set.

        References
        ----------
        - GraphicsBase skew_x and skew_y interfaces document
            - https://simon-ritchie.github.io/apysc/graphics_base_skew.html
        """
        with ap.DebugInfo(
                callable_='skew_x', locals_=locals(),
                module_name=__name__, class_=SkewXInterface):
            from apysc._validation import number_validation
            self._initialize_skew_x_if_not_initialized()
            number_validation.validate_integer(integer=value)
            before_value: ap.Int = self._skew_x
            self._skew_x = value
            self._append_skew_x_update_expression(before_value=before_value)

            self._append_skew_x_attr_linking_setting()

    def _append_skew_x_update_expression(self, before_value: ap.Int) -> None:
        """
        Append the skew x updating expression.

        Parameters
        ----------
        before_value : ap.Int
            Before updating value.
        """
        with ap.DebugInfo(
                callable_=self._append_skew_x_update_expression,
                locals_=locals(),
                module_name=__name__, class_=SkewXInterface):
            from apysc._type import value_util
            before_value_str: str = value_util.get_value_str_for_expression(
                value=before_value)
            after_value_str: str = value_util.get_value_str_for_expression(
                value=self._skew_x)
            expression: str = (
                f'{self.variable_name}.skew(-{before_value_str}, 0);'
                f'\n{self.variable_name}.skew({after_value_str}, 0);'
                f'\n{before_value_str} = {after_value_str};'
            )
            ap.append_js_expression(expression=expression)

    _skew_x_snapshots: Dict[str, int]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_skew_x_snapshots'):
            self._skew_x_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_skew_x_if_not_initialized()
        self._skew_x_snapshots[snapshot_name] = int(self._skew_x._value)

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._skew_x._value = self._skew_x_snapshots[snapshot_name]
